/***********************************************************

  OLED SSD 1306 (128x64) IIC-Interface library


License: LGPL V3
Copyright by Sven K. (oog / proog.de), 2018


This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, Or
(at your option) any later version.

This progr am is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


History:
2018-10-17: first public version

***********************************************************/


#include "oled_ssd1306.h"
#include <Wire.h> // I2C lib

//OLED::OLED() {
//}

void OLED::D1306_cmd(uint8_t c) {
  // I2C
  uint8_t control = 0x00;   // Co = 0, D/C = 0
  Wire.beginTransmission(_i2caddr);
  Wire.write(control);
  Wire.write(c);
  uint8_t r = Wire.endTransmission();
  if (r) {
    if (oledOk) {
      Serial.print(F(" IIC Error Addr=0x"));
      Serial.print(_i2caddr, 16);
      Serial.print(F(", Status="));
      Serial.println(r);
      oledOk = 0;
    }
  }
}

void OLED::D1306_cmd(uint8_t c, uint8_t d) {
  // I2C command with parameter in one I2C telegram
  uint8_t control = 0x00;   // Co = 0, D/C = 0
  Wire.beginTransmission(_i2caddr);
  Wire.write(control);
  Wire.write(c);
  Wire.write(d);
  uint8_t r = Wire.endTransmission();
  if (r) {
    if (oledOk) {
      Serial.print(F(" IIC Error Addr=0x"));
      Serial.print(_i2caddr, 16);
      Serial.print(F(", Status="));
      Serial.println(r);
      oledOk = 0;
    }
  }
}


// initialize OLED display

void OLED::D1306_begin(uint8_t i2caddr) {
  _i2caddr = i2caddr;

  //  Software Initialization Flow Chart

  D1306_cmd(D1306_DISPLAYON | 0);

  //  Enable charge pump regulator (8Dh, 14h)
  D1306_cmd(D1306_CHARGEPUMP, 0x14);

  //  Set MUX Ratio (A8h, 3Fh)
  D1306_cmd(D1306_MUXRATIO, D1306_HEIGHT - 1);

  //  Set Display Offset (D3h, 00h)
  D1306_cmd(D1306_DISPLAYOFFSET, 0x0);

  //  Set Display Start Line (40h)
  D1306_cmd(D1306_STARTLINE | 0x0);

  //  Set Segment re-map (A0h/A1h)
  D1306_cmd(D1306_SEGREMAP | 0x1);
  //D1306_cmd(D1306_SEGREMAP | 0x0);

  //  Set COM Output Scan Direction (C0h/C8h)
  //D1306_cmd(D1306_COMSCAN);
  D1306_cmd(D1306_COMSCANVERTFLIP);

  //  Set Contrast Control (81h, 7Fh)
  D1306_cmd(D1306_CONTRAST, 0x7F);

  //  Disable Entire Display On (A4h)
  D1306_cmd(D1306_ENTIREDISPON | 0);

  //  Set Normal Display (A6h)
  D1306_cmd(D1306_INVERSE | 0);

  //  Set Osc Frequency (D5h, 80h)
  D1306_cmd(D1306_DISPLAYCLKDIV, 0x80);

  //  Set COM Pins hardware configuration (DAh, 02)
  D1306_cmd(D1306_COMPINSHWCONFIG, 0x12); // 0x02

  //  Extra...
  D1306_cmd(D1306_MEMORYMODE, 0x00);
  //  D1306_cmd(D1306_DISPLAYCLKDIV, 0x80);
  //  D1306_cmd(D1306_PRECHARGEPERIOD, 0xF1);
  //  D1306_cmd(D1306_VCOMHADJUST, 0x40);

  //  Display On (AFh)
  D1306_cmd(D1306_DISPLAYON | 1);
}


void OLED::displayCursXY(uint8_t x, uint8_t y) {
  _displayCursX = x;
  _displayCursY = y;
  D1306_cmd(D1306_PGSTARTADDR | (y & 0x07));
  D1306_cmd(D1306_SETLOWCOLUMN | (x & 0x0f));
  D1306_cmd(D1306_SETHIGHCOLUMN | ((x >> 4) & 0x0f));
}


void OLED::displayFill(uint8_t d) {
  displayCursXY(0, 0);
  for (uint8_t i = 0; i < (D1306_WIDTH / 16 * D1306_HEIGHT / 8); i++) {
    Wire.beginTransmission(_i2caddr);
    Wire.write(0x40);
    for (uint8_t x = 0; x < 16; x++) {
      Wire.write(d);
    }
    Wire.endTransmission();
  }
}


void OLED::displayByte(uint8_t d) {
  Wire.beginTransmission(_i2caddr);
  Wire.write(0x40);
  Wire.write(d);
  Wire.endTransmission();
}


void OLED::displayContrast(uint8_t d) {
  D1306_cmd(D1306_CONTRAST, d);
}


/////////////////////////////////////////////////////
//
// Direct print to OLED
//
/////////////////////////////////////////////////////

void OLED::displayBuf(void) {
  Wire.beginTransmission(_i2caddr);
  Wire.write(0x40);
  for (uint16_t i = 1; i <= _buf16[0]; i++) {
    Wire.write(_buf16[i]);
  }
  Wire.endTransmission();
}


void OLED::displayBufN(uint8_t ptr[], uint8_t count) {
  Wire.beginTransmission(_i2caddr);
  Wire.write(0x40);
  for (uint16_t i = 0; i < count; i++) {
    Wire.write(ptr[i]);
  }
  Wire.endTransmission();
}


// display vertical stretched character
// _charStretchY
// 1: normal height, one line
// 2: double height, two lines (upper)
// 3: double height, two lines (lower)
// 4: quad height, four lines (upper)
// 5: quad height, four lines (upper mid)
// 6: quad height, four lines (lower mid)
// 7: quad height, four lines (lower)


void OLED::displayChar(uint8_t c) {
  uint8_t d;

  if ((c > 31) && (c < 127)) {
    const uint8_t* p = font + (c - 32) * fontCharacterBytes;
    // up to 16 bytes in a IIC tranmission
    Wire.beginTransmission(_i2caddr);
    Wire.write(0x40);
    for (uint8_t i = 0; i < fontCharacterBytes; i++) {
      d = _charInverse ^ pgm_read_byte_near(p++);
      switch (_charStretchY) {

        // line 1/1
        case 1:
          for (uint8_t i = 0; i < _charStretchX; i++) {
            Wire.write(d);
          }
          break;

        // line 1/2
        case 2:
          for (uint8_t i = 0; i < _charStretchX; i++) {
            Wire.write(stretch4bits[15 & d]);
          }
          break;

        // line 2/2
        case 3:
          for (uint8_t i = 0; i < _charStretchX; i++) {
            Wire.write(stretch4bits[d >> 4]);
          }
          break;

        // line 1/4
        case 4:
          for (uint8_t i = 0; i < _charStretchX; i++) {
            Wire.write(stretch2bits[d & 3]);
          }
          break;

        // line 2/4
        case 5:
          for (uint8_t i = 0; i < _charStretchX; i++) {
            Wire.write(stretch2bits[(d >> 2) & 3]);
          }
          break;

        // line 3/4
        case 6:
          for (uint8_t i = 0; i < _charStretchX; i++) {
            Wire.write(stretch2bits[(d >> 4) & 3]);
          }
          break;

        // line 4/4
        case 7:
          for (uint8_t i = 0; i < _charStretchX; i++) {
            Wire.write(stretch2bits[d >> 6]);
          }
          break;

      }
    }
    Wire.endTransmission();
  }
}


// display font size and format
// fntsize:
//   --6x8----8x8-- font
// 1:  6x 8 ( 8x 8)
// 2:  6x16 ( 8x16)
// 3: 12x16 (16x16)
// 4: 12x32 (16x32)
// 5: 24x32 (32x32)
//
// format:
// NORMAL
// INVERSE
// UNDERLINE
// INVERSE+UNDERLINE

void OLED::displayFormat(uint8_t fntsize, uint8_t format) {
  _charFontsize = fntsize;
  if (format & INVERSE) {
    _charInverse = 255;
  } else {
    _charInverse = 0;
  }
  if (format & UNDERLINE) {
    _charInverse ^= 128;
  }
}


void OLED::displayPrintXY(uint8_t x, uint8_t y, char s[]) {
  int i = 0;

  // fntsize=0, 1, 2, 3
  switch (_charFontsize) {
    case 1:
      _charStretchX = 1;
      _charStretchY = 1;
      displayCursXY(x, y);
      // print normal height, one line
      while (s[i] != 0) {
        displayChar(s[i++]);
      }
      break;

    case 2:
      _charStretchX = 1;
      _charStretchY = 2;
      displayCursXY(x, y);
      // print normal height, one line
      while (s[i] != 0) {
        displayChar(s[i++]);
      }
      _charStretchY = 3;
      displayCursXY(x, y + 1);
      // print normal height, one line
      i = 0;
      while (s[i] != 0) {
        displayChar(s[i++]);
      }
      break;

    case 3:
      _charStretchX = 2;
      _charStretchY = 2;
      displayCursXY(x, y);
      // print normal height, one line
      while (s[i] != 0) {
        displayChar(s[i++]);
      }
      _charStretchY = 3;
      displayCursXY(x, y + 1);
      // print normal height, one line
      i = 0;
      while (s[i] != 0) {
        displayChar(s[i++]);
      }
      break;

    case 4:
      _charStretchX = 2;
      _charStretchY = 4;
      displayCursXY(x, y);
      // print normal height, one line
      while (s[i] != 0) {
        displayChar(s[i++]);
      }
      _charStretchY = 5;
      displayCursXY(x, y + 1);
      // print normal height, one line
      i = 0;
      while (s[i] != 0) {
        displayChar(s[i++]);
      }
      _charStretchY = 6;
      displayCursXY(x, y + 2);
      // print normal height, one line
      i = 0;
      while (s[i] != 0) {
        displayChar(s[i++]);
      }
      _charStretchY = 7;
      displayCursXY(x, y + 3);
      // print normal height, one line
      i = 0;
      while (s[i] != 0) {
        displayChar(s[i++]);
      }
      break;

    case 5:
      _charStretchX = 4;
      _charStretchY = 4;
      displayCursXY(x, y);
      // print normal height, one line
      while (s[i] != 0) {
        displayChar(s[i++]);
      }
      _charStretchY = 5;
      displayCursXY(x, y + 1);
      // print normal height, one line
      i = 0;
      while (s[i] != 0) {
        displayChar(s[i++]);
      }
      _charStretchY = 6;
      displayCursXY(x, y + 2);
      // print normal height, one line
      i = 0;
      while (s[i] != 0) {
        displayChar(s[i++]);
      }
      _charStretchY = 7;
      displayCursXY(x, y + 3);
      // print normal height, one line
      i = 0;
      while (s[i] != 0) {
        displayChar(s[i++]);
      }
      break;
  }
}

void OLED::displayPrintXY(uint8_t x, uint8_t y, const __FlashStringHelper *s) {
  const int _charperline = D1306_WIDTH / fontCharacterBytes;
  char st[_charperline + 1];
  int i = 0;

  PGM_P p = reinterpret_cast<PGM_P>(s);
  while (i < _charperline) {
    st[i] = pgm_read_byte(p++);
    if (st[i] == 0) break;
    i++;
  }
  st[i] = 0;
  displayPrintXY(x, y, &st[0]);
}


void OLED::displayXYUInt(uint8_t x, uint8_t y, uint16_t i, uint8_t digits) {
  char digitstr[6] = "00000";
  if (digits < 1) return;
  if (digits > 5) return;
  digitstr[4] = '0' + ((i / 1) % 10);
  digitstr[3] = '0' + ((i / 10) % 10);
  digitstr[2] = '0' + ((i / 100) % 10);
  digitstr[1] = '0' + ((i / 1000) % 10);
  digitstr[0] = '0' + ((i / 10000) % 10);
  displayPrintXY(x, y, &digitstr[5] - digits);
}


void OLED::displayClrLine(uint8_t y) {
  uint8_t lMax = 0;
  switch (_charFontsize) {
    case 1:
      lMax = 1;
      break;
    case 2:
    case 3:
      lMax = 2;
      break;
    case 4:
    case 5:
      lMax = 4;
      break;
  }
  for (uint8_t l = 0; l < lMax; l++) {
    displayCursXY(0, y + l);
    //    Serial.print("Clear Line (y+l):");
    //    Serial.println(y+l);
    for (uint8_t i = 0; i < (D1306_WIDTH / 16); i++) {
      Wire.beginTransmission(_i2caddr);
      Wire.write(0x40);
      for (uint8_t x = 0; x < 16; x++) {
        Wire.write(0);
      }
      //      for (uint8_t x=0; x<8; x++) {
      //        Wire.write(0x55);
      //        Wire.write(0xAA);
      //      }
      Wire.endTransmission();
    }
  }
  //  displayCursXY(0,y);
}





/////////////////////////////////////////////////////
//
// Scrolltext Buffer
//
/////////////////////////////////////////////////////


#ifdef OLEDSSD1306_SCROLL

void OLED::initScreenBuffer() {
  for (int i = 0; i < (128 * 4); i++) {
    oledScreen[i] = 0;
  }
}


void OLED::displayLinebuf4(uint8_t y) {
  //for (uint8_t l = 0; l < 4; l++) {
  for (int8_t l = 3; l >= 0; l--) {
    displayCursXY(0, y + l);
    for (uint16_t j = 0; j < 8; j++) {
      Wire.beginTransmission(_i2caddr);
      Wire.write(0x40);
      for (uint16_t i = 0; i < 16; i++) {
        Wire.write(oledScreen[l * 128 + j * 16 + i]);
      }
      Wire.endTransmission();
    }
  }
}


void OLED::displayChar2Buf(uint8_t c, int16_t pos) {
  uint8_t d;
  uint16_t p;

  if ((c > 31) && (c < 127)) {
    const uint8_t* p = font + (c - 32) * fontCharacterBytes;
    for (uint8_t j = 0; j < fontCharacterBytes; j++) {
      d = _charInverse ^ pgm_read_byte_near(p++);

      // line 1/4
      for (uint8_t i = 0; i < _charStretchX; i++) {
        if (((pos + i + j * _charStretchX) >= 0) && ((pos + i + j * _charStretchX) < 128)) {
          // line 1/4
          oledScreen[pos + 0 * 128 + i + j * _charStretchX] = stretch2bits[d & 3];
          // line 2/4
          oledScreen[pos + 1 * 128 + i + j * _charStretchX] = stretch2bits[(d >> 2) & 3];
          // line 3/4
          oledScreen[pos + 2 * 128 + i + j * _charStretchX] = stretch2bits[(d >> 4) & 3];
          // line 4/4
          oledScreen[pos + 3 * 128 + i + j * _charStretchX] = stretch2bits[(d >> 6) & 3];
        }
      }
    }
  }
}


void OLED::displayText2Buf(char t[], int16_t pos) {
  int16_t tempPos = pos;
  for (int i = 0; t[i] != 0; i++) {
    displayChar2Buf(t[i], tempPos);
    tempPos += (_charStretchX * fontCharacterBytes);
  }
  // add blank character
//  displayChar2Buf(' ', tempPos);
}


void OLED::scrollText(char* s, uint8_t y) {
  uint16_t w = _charStretchX * fontCharacterBytes * strlen(s);
  for (int i = D1306_WIDTH; i > -(w); i=i-2) {
    displayText2Buf(s, i);
    displayLinebuf4(y);
    delay(1);
  }
}


void OLED::scrollTextFast(char* s, uint8_t y) {
  uint16_t w = _charStretchX * fontCharacterBytes * strlen(s);
  for (int i = D1306_WIDTH; i > -(w); i=i-3) {
    displayText2Buf(s, i);
    displayLinebuf4(y);
    if((i&1) == 0) delay(1);
  }
}

#endif

